//
// tbbpipecrypto - tbbpipecrypto.cpp
//
#include <stdio.h>
#include <iostream>
#include <cstring>
#include <cctype>
#include "tbb/task_scheduler_init.h"
#include "tbb/pipeline.h"

using namespace std;
using namespace tbb;

const int BUFFER_SIZE = 2048;

class MyBuffer {
    static const size_t buffer_size = BUFFER_SIZE;
    char* my_end;
    char storage[buffer_size];
public:
    char* begin() {return storage;}
    const char* begin() const {return storage;}
    char* end() const {return my_end;}
    void set_end( char* new_ptr ) {my_end = new_ptr;}
    size_t max_size() const {return buffer_size;}
    size_t size() const {return my_end-begin();}
};

class InputFilter: public tbb::filter {
public:
    static const size_t n_buffer = 4;
private:
    FILE* input_file;
    MyBuffer buffer[n_buffer];
public:
    InputFilter( FILE* fp ) : filter(true), input_file(fp) {}
    void* operator()(void*) {
        MyBuffer& b = buffer[0];
        size_t n = fread( b.begin(), 1, b.max_size(), input_file );
        if( n == 0 ) {
            return NULL;
        } else {
            b.set_end( b.begin()+n );
            return &b;
        }
    }
};

char Caesar(char s)
{
    int c = (int)s;
    if (!isalnum(c)) {
        return s;
    }
    int cc = c + 3;
    if (isdigit(c)!=0 & (cc>0x39))
        cc -= 10;
    if (isupper(c)!=0 & (cc>0x5a))
        cc -= 26;
    if (islower(c)!=0 & (cc>0x7a))
        cc -= 26;
    return (char)cc;
}

class CryptoFilter: public tbb::filter {
public:
    CryptoFilter() : tbb::filter(/*serial=*/false){}
    void* operator()( void* item ) {
        MyBuffer& b = *static_cast<MyBuffer*>(item);
        for( char* s=b.begin(); s!=b.end(); ++s ) {
            *s = Caesar(*s);
        }
        return &b;
    }
};

class OutputFilter: public tbb::filter {
    FILE* output_file;
public:
    OutputFilter( FILE* fp ) : tbb::filter(true), output_file(fp) {}
    void* operator()( void* item ) {
        MyBuffer& b = *static_cast<MyBuffer*>(item);
        fwrite( b.begin(), 1, b.size(), output_file );
        return NULL;
    }
};

char *decrypto(char *text)
{
    char buffer[BUFFER_SIZE];
    size_t len = strlen(text);
    for (size_t i=0; i<len; i++) {
        int c = (int)text[i];
        if (!isalnum(c)) {
            buffer[i] = text[i];
        }else {
            int cc = c - 3;
            if (isdigit(c)!=0 & (cc<0x30))
                cc += 10;
            if (isupper(c)!=0 & (cc<0x41))
                    cc += 26;
            if (islower(c)!=0 & (cc<0x61))
                cc += 26;
            buffer[i] = (char)cc;
        }
    }
    buffer[len] = '\0';
    strcpy(text, buffer);
    return text;
}

int main(int argc, char * argv[])
{
    task_scheduler_init init;
    
    FILE *fpi = fopen("plain.txt", "r");
    if (!fpi) {
        printf("plain.txt open error\n");
        return -1;
    }

    FILE *fpo = fopen("cipher.dat", "w");
    if (!fpo) {
        printf("cipher.dat open error\n");
        return -1;
    }

    tbb::pipeline pipeline;
    // input
    InputFilter input_filter( fpi );
    pipeline.add_filter( input_filter );
    // cryption
    CryptoFilter transform_filter1;
    pipeline.add_filter( transform_filter1 );
    // output
    OutputFilter output_filter( fpo );
    pipeline.add_filter( output_filter );
    // Run the pipeline
    pipeline.run( InputFilter::n_buffer );
    pipeline.clear();
    fclose(fpi);
    fclose(fpo);

    // decryption
    char TmpBuffer[BUFFER_SIZE];
    int count = 0;
    FILE *fpr = fopen("cipher.dat", "r");
    int ch;
    while ((ch = fgetc(fpr))!= EOF){
        TmpBuffer[count++] = ch;
    }
    fclose(fpr);
    TmpBuffer[count] = '\0';

    printf("plain=\n%s\n", decrypto(TmpBuffer));

    return 0;
}
